# Basic Information
__name__ = "pipemake"
__version__ = "1.1.6"
__summary__ = "Pipemake: A pipeline creation tool using Snakemake"
__url__ = "https://github.com/kocherlab/pipemake"
__code__ = "https://github.com/kocherlab/pipemake"
__issue__ = "https://github.com/kocherlab/pipemake/issues"
__docs__ = "https://pipemake.readthedocs.io/en/latest/"
__license__ = "MIT"
__copyright__ = "2023"

# Author Information
__authors__ = "Andrew Webb, Scott Wolf, Ian Traniello, and Sarah Kocher"
__email__ = "19213578+aewebb80@users.noreply.github.com"
